//         ***************************
//         * 24h-Temperatur-Recorder *
//         ***************************

// Pinbelegung:
// ============
// PORTC   7     6    5    4    3    2    1    0
//         -     -   SCL  SDA  n.u. n.u. n.u. n.u.
// ACHTUNG: SCL und SDA benötigen pullup-Widerstände (2..5kOhm) nach +5V

// PORTD Display-Data

// PORTB   7     6    5    4    3    2    1    0
//         -     -   n.c. CS1  CS2   E   RES  RS(=D/I)


// globale Variablen
// -----------------
float temp, temp2;       // die gemessene Temperatur, Zwischenspeicher
unsigned int temp_int;   // integer Wert (Absolutwert)
unsigned char measval=0; // Meßwert

unsigned char num[11][5]=
              {{ 62, 81, 73, 69, 62 },  // 0
               {  0, 66,127, 64,  0 },  // 1
               { 98, 81, 73, 69, 66 },  // 2
               { 33, 65, 69, 75, 49 },  // 3
               { 24, 20, 18,127, 16 },  // 4
               { 39, 69, 69, 69, 57 },  // 5              
               { 60, 74, 73, 73, 48 },  // 6
               {  1,113,  9,  5,  3 },  // 7
               { 54, 73, 73, 73, 54 },  // 8
               {  6, 73, 73, 41, 30 },  // 9
               {  0,  0,  8,  8,  8 }}; // -

unsigned char dmask[8][128]; // Display-Register-Maske (page, x) 
unsigned char temph, templ;  // Highbyte und Lowbyte der gemessenen Temperatur
unsigned char meas_flag=1;   // 0=nächste Messung abwarten, 1=Messung ausführen
unsigned int  count_sec=0;   // Counter für die Sekunden bis zur nächsten Messungen

// *****************************************************************************************
// Grafikdisplay

void ChipSelect1 (void) // set CS1=1, CS2=0
{     
 PORTB |= 0x10; // CS1=1 => die linke Hälfte des Displays
 PORTB &= 0xF7; // CS2=0
 }


void ChipSelect2 (void) // set CS1=0, CS2=1
{     
 PORTB |= 0x08; // CS2=1; // die rechte Hälfte des Displays
 PORTB &= 0xEF; // CS1=0;
 }

  
void CLK_E (void) // E clocken
{   
 unsigned int i;
  
 PORTB &= 0xFB; // E=0
 PORTB |= 0x04; // E=1; 
 delayMicroseconds(10); // ab 6 funktioniert das Display richtig
 PORTB &= 0xFB; // E=0
 }

// --------------------------------------------------------------------------

void Write_CData (unsigned char cs) // write Control Data to CS1 or CS2
{     
 if (cs==1) ChipSelect1(); // das entsprechende Chip-Select setzen
 if (cs==2) ChipSelect2();
 PORTB &= 0xFE; // RS=0 
 CLK_E();
 }


void Write_DData (unsigned char cs) // write Display RAM Data
{   
 if (cs==1) ChipSelect1(); // das entsprechende Chip-Select setzen
 if (cs==2) ChipSelect2();
 PORTB |= 0x01; // RS=1
 CLK_E();
 }


void Ini_Gdisp (void) // das Display initialisieren
{ 
 PORTB &= 0xE0; // CS1=CS21=E=RES=RS=0   
 delay(100);    
 PORTB |= 0x02; // RES=1 (Reset inactive)
 delay(100);        
 }


void clear_dispreg (void)     // Display-Register löschen
{
 unsigned char page,x;        // Seite, Spalte

 for (page=0; page<8; page++) // alle Meßwerte nach links schieben
 {
  for (x=0; x<128; x++) dmask[page][x]=0x00;
  }
 }


void set_ytics (void)     // tics setzen
{
 unsigned char x;        // Spalte
   
 for (x=16; x<18; x++) dmask[0][x]=1;  //  40° linke Seite
 for (x=16; x<18; x++) dmask[1][x]=4;  //  30°
 for (x=16; x<18; x++) dmask[2][x]=16; //  20°
 for (x=16; x<18; x++) dmask[3][x]=64; //  10°
 for (x=16; x<18; x++) dmask[6][x]=4;  // -10°
 for (x=16; x<18; x++) dmask[7][x]=16; // -20°  
 }

 
void set_ylabels (void) // y-labels setzen
{ 
 unsigned char x;        // Spalte

 dmask[0][1]=0;   // ° (der obere Teil)
 dmask[0][2]=0;  
 dmask[0][3]=48;
 dmask[0][4]=48; 
 dmask[0][5]=0;   
 dmask[0][7]=240; // C 
 dmask[0][8]=8;  
 dmask[0][9]=8;  
 dmask[0][10]=8; 
 dmask[0][11]=16;

 dmask[1][1]=0;   // (der untere Teil)
 dmask[1][2]=0;  
 dmask[1][3]=0;
 dmask[1][4]=0; 
 dmask[1][5]=0;   
 dmask[1][7]=1;  
 dmask[1][8]=2;  
 dmask[1][9]=2;  
 dmask[1][10]=2; 
 dmask[1][11]=1;
 // ---------------------------------------
 dmask[2][3]=196;  // 2 (+20°)
 dmask[2][4]=162;  
 dmask[2][5]=146; 
 dmask[2][6]=138;  
 dmask[2][7]=132;   
 dmask[2][9]=124;  // 0 
 dmask[2][10]=162;  
 dmask[2][11]=146;  
 dmask[2][12]=138;  
 dmask[2][13]=124;
 // ---------------------------------------
 dmask[3][3]=0;    // 1 (+10°, der obere Teil)
 dmask[3][4]=16;  
 dmask[3][5]=248; 
 dmask[3][6]=0;  
 dmask[3][7]=0;   
 dmask[3][9]=240;  // 0 
 dmask[3][10]=136;  
 dmask[3][11]=72;  
 dmask[3][12]=40;  
 dmask[3][13]=240;

 dmask[4][3]=0;    // (der untere Teil)
 dmask[4][4]=2;  
 dmask[4][5]=3; 
 dmask[4][6]=2;  
 dmask[4][7]=0;   
 dmask[4][9]=1;    
 dmask[4][10]=2;  
 dmask[4][11]=2; 
 dmask[4][12]=2;  
 dmask[4][13]=1;
 // ---------------------------------------
 dmask[6][0]=4;   // -
 dmask[6][1]=4;
 dmask[6][2]=4;

 dmask[5][3]=0;   // 1 (-10°, der obere Teil)
 dmask[5][4]=0;
 dmask[5][5]=128; 
 dmask[5][6]=0;  
 dmask[5][7]=0;   
 dmask[5][9]=0;   // 0 
 dmask[5][10]=128;  
 dmask[5][11]=128;  
 dmask[5][12]=128;  
 dmask[5][13]=0;  
 
 dmask[6][3]=0;   // (der untere Teil)
 dmask[6][4]=33;
 dmask[6][5]=63; 
 dmask[6][6]=32;  
 dmask[6][7]=0;   
 dmask[6][9]=31;  
 dmask[6][10]=40;  
 dmask[6][11]=36;  
 dmask[6][12]=34;  
 dmask[6][13]=31;  
 // ---------------------------------------
 dmask[7][0]=16;   // -
 dmask[7][1]=16;
 dmask[7][2]=16;

 dmask[7][3]=196;  // 2 (-20°)
 dmask[7][4]=162;  
 dmask[7][5]=146; 
 dmask[7][6]=138;  
 dmask[7][7]=132;   
 dmask[7][9]=124;  // 0 
 dmask[7][10]=162;  
 dmask[7][11]=146;  
 dmask[7][12]=138;  
 dmask[7][13]=124;

 for (x=0; x<128; x++) dmask[5][x]=dmask[5][x] | 0x01; // Nulllinie (y=40-Temp)  
 for (x=0; x<8; x++) dmask[x][18]=255;                 // y-Linie  
 }


void shift_left (void)        // alle Meßwerte nach links schieben
{
 unsigned char page,x;        // Seite, Spalte

 for (page=0; page<8; page++) // alle Meßwerte nach links schieben
 {
  for (x=19; x<127; x++) dmask[page][x]=dmask[page][x+1]; // ein Messwert pro 800s über 24h
  }
 }


void set_meas (unsigned char y)     // einen neuen Meßwert im Display-Register speichern
{
 unsigned char page;  // dem Meßwert entsprechende Seite (0..7)  
 unsigned char line;  // Zeile in dieser Seite (0..7)
 unsigned char ddata; // entsprechender Wert (1..128)

 for (page=0; page<8; page++) dmask[page][127]=0x00;   // die ganze letzte Spalte löschen   
 
 page=y/8;            
 line=y-8*page;      

 ddata=1; 
 ddata=ddata << line;              

 dmask[page][127]=ddata;  

 dmask[5][127]=dmask[5][127] | 1; // Nullinie nicht löschen (y=40-Temp)
 }


void load_gdisp() // das komplette Grafikdisplay mit dem Display-Register laden
{    
 unsigned char cs;                 // chip select
 unsigned char page;               // Seite (0...7)    
 unsigned char x;                  // Zeile
 unsigned char ym;                 // Maske für die Y-address

 for (page=0; page<8; page++)      // alle Seiten durchlaufen
 {
  for (x=0; x<128; x++) 
  {
   if (x<64)                       // set Y-address (Spalte) in der Display-Hälfte
   {
    cs=1;                          // chip select 1
    ym = 0x40 | x;                 // x=0..63, ym=0x40..0x7F
    }

   else
   { 
    cs=2;                          // chip select 2
    ym = 0x40 | (x-64);            // x=64..127, ym=0x40..0x7F
    }               

   PORTD  =  0xC0;                 // start line=0 
   Write_CData(cs);    

   PORTD=ym;                       // die Adresse laden
   Write_CData(cs);

   PORTD=0xB8 | page;              // set Page 
   Write_CData(cs); 
   
   PORTD=dmask[page][x];           // das neue Datum laden
   Write_DData(cs);                // write Display RAM Data 

   } // alle Spalten der aktuellen page geladen
  } // alle pages geladen             
   
  PORTD=0x3F;                      // Display on (Wiederholung ist notwendig !)
  Write_CData(1);                  // write Control Data     
  Write_CData(2);                  // write Control Data  
 }


void load_num (void) // die aktuelle Temperatur als Zahl rechts dazuschreiben
{     
 unsigned char x;                 // Zeile
 unsigned char ym;                // Maske für die Y-address
 unsigned char z10;               // Zehnerziffer
 unsigned char z1;                // Einerziffer
 unsigned char page=0;            // aktuelle zu testende Seite   
 unsigned char page_test=1;       // 1=testen, ob diese Seite frei ist
 
 while (page_test==1)             
 {
  page_test=0;                    // wenn kein schwarzes Pixel gefunden wid, kann diese Seite benutzt werden 
  for (x=111; x<128; x++)         // checken, ob der Ziffernraum belegt ist
  {
   if (dmask[page][x] != 0) page_test=1; // ist leider belegt
   }
  if (page_test==1) page++;       // nächste Seite testen
  }

 z10=temp_int/10;         
 z1=temp_int-10*z10;

 PORTD=0xB8 | page;               // set Page 
 Write_CData(2); 

 PORTD  =  0xC0;                  // start line=0 
 Write_CData(2);    

 for (x=111; x<128; x++)          // den gesamten Zahlenbereich im Display freimachen
 {
  ym = 0x40 | (x-64);             // x=64..127, ym=0x40..0x7F            
  PORTD=ym;                       // Spaltenadresse laden
  Write_CData(2);       
  PORTD=0;                        // blank
  Write_DData(2);                 // write Display RAM Data 
  }  
 // ---------------------------------------------------------------------------
 for (x=123; x<128; x++)          // die Einerziffer 
 {    
  ym = 0x40 | (x-64);                      
  PORTD=ym;                        
  Write_CData(2);      
  PORTD=num[z1][x-123];            
  Write_DData(2);             
  }    
 // ---------------------------------------------------------------------------
 if (z10>0)                       // die Zehnerziffer (führende Null blanken)
 {
  for (x=117; x<122; x++) 
  {   
   ym = 0x40 | (x-64);            
   PORTD=ym;                         
   Write_CData(2);      
   PORTD=num[z10][x-117];  
   Write_DData(2);              
   } 
  }
 // ---------------------------------------------------------------------------
 if (temp<=-0.5)                      // bei Minustemperatur
 //if (temp<=-1)                      // bei Minustemperatur
 {
  if (z10>0)                      // Temperaturwert zweistellig
  {
   for (x=111; x<116; x++) 
   {    
    ym = 0x40 | (x-64);           
    PORTD=ym;                         
    Write_CData(2);      
    PORTD=num[10][x-111];         // Minuszeichen links neben Zehnerziffer
    Write_DData(2);               
    } 
   }

  else                            // Temperaturwert einstellig
  {
   for (x=117; x<122; x++) 
   {    
    ym = 0x40 | (x-64);           
    PORTD=ym;                         
    Write_CData(2);      
    PORTD=num[10][x-117];         // Minuszeichen anstelle der Zehnerziffer
    Write_DData(2);               
    } 
   }
  }
 // ---------------------------------------------------------------------------           
  PORTD  =  0x3F;                 // Display on (Wiederholung ist notwendig !)
  Write_CData(2);                 // write Control Data  
 }
 
// ****************************************************************************

void check_TWINT (void) // warten bis TWINT gesetzt ist
{
 while ((TWCR&0x80)==0);
 //while (!(TWCR & (1<<TWINT))); // alternative Schreibweise
 }


void write_to_config(void) // Configuration Register laden
{
 TWCR=0xA4; // start condition
 check_TWINT();
 
 TWDR=0x90; // slave address and write
 TWCR=0x84; // start transmission
 check_TWINT();
  
 TWDR=0xAC; // master sends access config command
 TWCR=0x84; 
 check_TWINT();

 TWDR=0x0C; // 12-bit, continuous mode
 TWCR=0x84; 
 check_TWINT();
 
 TWCR=0x94; // stop condition
 }

void start_convert (void) // die Temperatur-Konvertierung starten
{
 TWCR=0xA4; // start condition
 check_TWINT();

 TWDR=0x90; // slave address and write
 TWCR=0x84; // start transmission
 check_TWINT();
 
 TWDR=0x51; // start convert command
 TWCR=0x84; 
 check_TWINT();

 TWCR=0x94; // stop condition
 }


void read_temp (void) // Die Temperatur vom DS1631 einlesen
{  
 TWCR=0xA4; // start condition
 check_TWINT();

 TWDR=0x90; // slave address and write
 TWCR=0x84; // start transmission
 check_TWINT();
 
 TWDR=0xAA; // read temperature 
 TWCR=0x84; 
 check_TWINT();

 TWCR=0xA4; // repeat start condition
 check_TWINT();

 TWDR=0x91; // slave address and read
 TWCR=0x84; // start transmission
 check_TWINT();

 TWCR=0x84; // read Highbyte
 check_TWINT();
 temph=TWDR;

 TWCR=0x84; // read Lowbyte
 check_TWINT();
 templ=TWDR;

 TWCR=0x94; // stop condition
 }
 

void cal_temp (void) // die Temperaturwert als float berechnen
{   
 if ((temph&0x80)==0) temp=0; else temp=-128;
 if ((temph&0x40)!=0) temp=temp+64;
 if ((temph&0x20)!=0) temp=temp+32;
 if ((temph&0x10)!=0) temp=temp+16;
 if ((temph&0x08)!=0) temp=temp+8;
 if ((temph&0x04)!=0) temp=temp+4;
 if ((temph&0x02)!=0) temp=temp+2;
 if ((temph&0x01)!=0) temp=temp+1;
 
 if ((templ&0x80)!=0) temp=temp+0.5;
 if ((templ&0x40)!=0) temp=temp+0.25;
 if ((templ&0x20)!=0) temp=temp+0.125;
 if ((templ&0x10)!=0) temp=temp+0.0625;
 }


void cal_measval (void) // Messwert bestimmen (-23°C => 63, +40°C => 0)
{
 if (temp<0)
 {
  temp_int=(0.5-temp);  // Temperaturwert aufrunden
  measval=40+temp_int;
  }

 else // Temperatur >=0
 {
  temp_int=temp+0.5;    // Temperaturwert aufrunden
  measval=40-temp_int;
  }
 }

// *********************************************************************************************************

void setup()
{
 DDRB=0xFF;         // alle Pins von PORTB push-pull 
 DDRD=0xFF;         // alle Pins von PORTD push-pull

 TWSR=0xFA;         // two wire prescaler :16
 TWBR=0x05;         // divider :5 => SCL-clock 91kHz
 
 write_to_config(); // Configuration Register des Thermometers DS1631 laden
 start_convert();   // die Temperatur-Konvertierung starten
 Ini_Gdisp();       // das Grafikdisplay 128x64 initialisieren
 clear_dispreg();   // das Display-Register löschen
 set_ytics();       // y-tics setzen
 set_ylabels();     // y-labels setzen
 
 load_gdisp();      // das komplette Grafikdisplay mit dem frame laden
 delay(1000);       // warten bis der DS1631 bereit ist  

 TCCR1A=0x00;       // normal mode, ACHTUNG: Initial Value wurde überschrieben !
 TCCR1B=0x04;       // Timer 1 clock prescaler :256
 TIMSK1=0x01;       // Timer 1 overflow interrupt enabled
 SREG=0x80;         // global interrupt enabled
}
 
void loop()
{      
 if (meas_flag==1)      // einen Temperaturwert messen und darstellen 
 {
  meas_flag=0;          // Wartezeit bis zur nächsten Messung starten              
 
  read_temp();          // die Temperatur vom DS1631 einlesen
  cal_temp();           // die Temperatur berechnen
  temp2=temp;           // save temp
 
  if (temp>40)  temp=0; // Bereichsüberschreitung => auf Nulllinie umleiten
  if (temp<-23) temp=0; //                  "
  cal_measval();        // den y-Wert des Displays berechnen
  set_meas(measval);    // einen neuen Meßwert setzen  
  load_gdisp();         // das komplette Grafikdisplay mit dem Display-Register laden   

  temp=temp2;           // restore temp    
  cal_measval();           // nochmals unlimitierte Temperatur aufrunden 
  load_num();           // die aktuelle Temperatur als Zahl im Display einblenden
  //delay(500);           // 60s/Durchlauf         
  //delay(5000);          // 9min/Durchlauf 
  //delay(32860);         // 1h/Durchlauf
  //delay(789000);        // 24h/Durchlauf
  shift_left();         // alle alten Meßwerte nach links schieben   
  } // Ende der aktuellen messung 
 }  


ISR(TIMER1_OVF_vect) 
{  
 TCNT1=0x0BDC;       // 1s*16MHz/256=62500counts bis 65536 => Startwert 3036
 count_sec++;        // Sekundenzähler
 //if (count_sec==3) // die nächste Messung starten (test only)
 if (count_sec==800) // Messintervall 24*3600s/108=800s abgelaufen => nächste Messung starten
 {
  count_sec=0;
  meas_flag=1;       // die Messung auslösen      
  }
 }
