//         ********************************
//         * 24-hour temperature recorder *
//         ********************************

// Pin assignment:
// ===============
// PORTC   7     6    5    4    3    2    1    0
//         -     -   SCL  SDA  nc   nc   nc   nc
// ATTENTION: SCL and SDA need pullup resistors (2..5kOhm) to +5V

// PORTD Display Data

// PORTB   7     6    5    4    3    2    1    0
//         -     -   n.c. CS1  CS2   E   RES  RS(=D/I)


// global variables
// -----------------
float temp, temp2;       // measured temperature, buffer memory
unsigned int temp_int;   // integer value (absolute value)
unsigned char measval=0; // measured value

unsigned char num[11][5]=
              {{ 62, 81, 73, 69, 62 },  // 0
               {  0, 66,127, 64,  0 },  // 1
               { 98, 81, 73, 69, 66 },  // 2
               { 33, 65, 69, 75, 49 },  // 3
               { 24, 20, 18,127, 16 },  // 4
               { 39, 69, 69, 69, 57 },  // 5              
               { 60, 74, 73, 73, 48 },  // 6
               {  1,113,  9,  5,  3 },  // 7
               { 54, 73, 73, 73, 54 },  // 8
               {  6, 73, 73, 41, 30 },  // 9
               {  0,  0,  8,  8,  8 }}; // -

unsigned char dmask[8][128]; // Display register mask (page, x) 
unsigned char temph, templ;  // Highbyte and lowbyte of the measured temperature
unsigned char meas_flag=1;   // 0=wait for next measurement, 1=execute measurement
unsigned int  count_sec=0;   // Count the seconds until the next measurement

// *****************************************************************************************
// Graphic display

void ChipSelect1 (void) // set CS1=1, CS2=0
{     
 PORTB |= 0x10; // CS1=1 => left part of display
 PORTB &= 0xF7; // CS2=0
 }


void ChipSelect2 (void) // set CS1=0, CS2=1
{     
 PORTB |= 0x08; // CS2=1; // right part of display
 PORTB &= 0xEF; // CS1=0;
 }

  
void CLK_E (void) // clock E
{   
 unsigned int i;
  
 PORTB &= 0xFB; // E=0
 PORTB |= 0x04; // E=1; 
 delayMicroseconds(10); // from 6us the display works correctly
 PORTB &= 0xFB; // E=0
 }

// --------------------------------------------------------------------------

void Write_CData (unsigned char cs) // write Control Data to CS1 or CS2
{     
 if (cs==1) ChipSelect1(); // set the corresponding chip select
 if (cs==2) ChipSelect2();
 PORTB &= 0xFE; // RS=0 
 CLK_E();
 }


void Write_DData (unsigned char cs) // write Display RAM Data
{   
 if (cs==1) ChipSelect1(); // set the corresponding chip select
 if (cs==2) ChipSelect2();
 PORTB |= 0x01; // RS=1
 CLK_E();
 }


void Ini_Gdisp (void) // dinitialize the display
{ 
 PORTB &= 0xE0; // CS1=CS21=E=RES=RS=0   
 delay(100);    
 PORTB |= 0x02; // RES=1 (Reset inactive)
 delay(100);        
 }


void clear_dispreg (void)     // Clear display register
{
 unsigned char page,x;        // Page, column

 for (page=0; page<8; page++) // move all measured values to the left
 {
  for (x=0; x<128; x++) dmask[page][x]=0x00;
  }
 }


void set_ytics (void)     // set tics
{
 unsigned char x;        // column
   
 for (x=16; x<18; x++) dmask[0][x]=1;  //  40° left side
 for (x=16; x<18; x++) dmask[1][x]=4;  //  30°
 for (x=16; x<18; x++) dmask[2][x]=16; //  20°
 for (x=16; x<18; x++) dmask[3][x]=64; //  10°
 for (x=16; x<18; x++) dmask[6][x]=4;  // -10°
 for (x=16; x<18; x++) dmask[7][x]=16; // -20°  
 }

 
void set_ylabels (void) // set y labels
{ 
 unsigned char x;        // column

 dmask[0][1]=0;   // ° (the upper part)
 dmask[0][2]=0;  
 dmask[0][3]=48;
 dmask[0][4]=48; 
 dmask[0][5]=0;   
 dmask[0][7]=240; // C 
 dmask[0][8]=8;  
 dmask[0][9]=8;  
 dmask[0][10]=8; 
 dmask[0][11]=16;

 dmask[1][1]=0;   // (the lower part)
 dmask[1][2]=0;  
 dmask[1][3]=0;
 dmask[1][4]=0; 
 dmask[1][5]=0;   
 dmask[1][7]=1;  
 dmask[1][8]=2;  
 dmask[1][9]=2;  
 dmask[1][10]=2; 
 dmask[1][11]=1;
 // ---------------------------------------
 dmask[2][3]=196;  // 2 (+20°)
 dmask[2][4]=162;  
 dmask[2][5]=146; 
 dmask[2][6]=138;  
 dmask[2][7]=132;   
 dmask[2][9]=124;  // 0 
 dmask[2][10]=162;  
 dmask[2][11]=146;  
 dmask[2][12]=138;  
 dmask[2][13]=124;
 // ---------------------------------------
 dmask[3][3]=0;    // 1 (+10°, the upper part)
 dmask[3][4]=16;  
 dmask[3][5]=248; 
 dmask[3][6]=0;  
 dmask[3][7]=0;   
 dmask[3][9]=240;  // 0 
 dmask[3][10]=136;  
 dmask[3][11]=72;  
 dmask[3][12]=40;  
 dmask[3][13]=240;

 dmask[4][3]=0;    // (the lower part)
 dmask[4][4]=2;  
 dmask[4][5]=3; 
 dmask[4][6]=2;  
 dmask[4][7]=0;   
 dmask[4][9]=1;    
 dmask[4][10]=2;  
 dmask[4][11]=2; 
 dmask[4][12]=2;  
 dmask[4][13]=1;
 // ---------------------------------------
 dmask[6][0]=4;   // -
 dmask[6][1]=4;
 dmask[6][2]=4;

 dmask[5][3]=0;   // 1 (-10°, the upper part)
 dmask[5][4]=0;
 dmask[5][5]=128; 
 dmask[5][6]=0;  
 dmask[5][7]=0;   
 dmask[5][9]=0;   // 0 
 dmask[5][10]=128;  
 dmask[5][11]=128;  
 dmask[5][12]=128;  
 dmask[5][13]=0;  
 
 dmask[6][3]=0;   // (the lower part)
 dmask[6][4]=33;
 dmask[6][5]=63; 
 dmask[6][6]=32;  
 dmask[6][7]=0;   
 dmask[6][9]=31;  
 dmask[6][10]=40;  
 dmask[6][11]=36;  
 dmask[6][12]=34;  
 dmask[6][13]=31;  
 // ---------------------------------------
 dmask[7][0]=16;   // -
 dmask[7][1]=16;
 dmask[7][2]=16;

 dmask[7][3]=196;  // 2 (-20°)
 dmask[7][4]=162;  
 dmask[7][5]=146; 
 dmask[7][6]=138;  
 dmask[7][7]=132;   
 dmask[7][9]=124;  // 0 
 dmask[7][10]=162;  
 dmask[7][11]=146;  
 dmask[7][12]=138;  
 dmask[7][13]=124;

 for (x=0; x<128; x++) dmask[5][x]=dmask[5][x] | 0x01; // Zero line (y=40-Temp)  
 for (x=0; x<8; x++) dmask[x][18]=255;                 // y line  
 }


void shift_left (void)        // shift all measured values to the left

 unsigned char page,x;        // page. column

 for (page=0; page<8; page++) // shift all measured values to the left
 {
  for (x=19; x<127; x++) dmask[page][x]=dmask[page][x+1]; // one measured value per 800s over 24h
  }
 }


void set_meas (unsigned char y)     // save a new measured value in the display register
{
 unsigned char page;  // page corresponding to the measured value (0..7)
 unsigned char line;  // Line in this page  (0..7)
 unsigned char ddata; // orresponding value (1..128)

 for (page=0; page<8; page++) dmask[page][127]=0x00;   // clear the whole last column
 
 page=y/8;            
 line=y-8*page;      

 ddata=1; 
 ddata=ddata << line;              

 dmask[page][127]=ddata;  

 dmask[5][127]=dmask[5][127] | 1; // Do not clear zero line (y=40-Temp)
 }


void load_gdisp() // load the complete graphic display with the display register
{    
 unsigned char cs;                 // chip select
 unsigned char page;               // page (0...7)    
 unsigned char x;                  // line
 unsigned char ym;                 // Maske for Y address

 for (page=0; page<8; page++)      // scroll through all pages
 {
  for (x=0; x<128; x++) 
  {
   if (x<64)                       // set Y address (column) in part of the display
   {
    cs=1;                          // chip select 1
    ym = 0x40 | x;                 // x=0..63, ym=0x40..0x7F
    }

   else
   { 
    cs=2;                          // chip select 2
    ym = 0x40 | (x-64);            // x=64..127, ym=0x40..0x7F
    }               

   PORTD  =  0xC0;                 // start line=0 
   Write_CData(cs);    

   PORTD=ym;                       // load addresse
   Write_CData(cs);

   PORTD=0xB8 | page;              // set page 
   Write_CData(cs); 
   
   PORTD=dmask[page][x];           // load new data
   Write_DData(cs);                // write display RAM data 

   } // all columns of the current page loaded
  } // all pages loaded              
   
  PORTD=0x3F;                      // Display on (repetition is necessary !)
  Write_CData(1);                  // write control data     
  Write_CData(2);                  // write control data  
 }


void load_num (void) // add the current temperature as a number to the right
{     
 unsigned char x;                 // line
 unsigned char ym;                // maske for Y address
 unsigned char z10;               // tens digit
 unsigned char z1;                // ones digit
 unsigned char page=0;            // current page to test   
 unsigned char page_test=1;       // 1=test if this page is free
 
 while (page_test==1)             
 {
  page_test=0;                    // if no black pixel is found, this page can be used
  for (x=111; x<128; x++)         // check if the digit space is occupied
  {
   if (dmask[page][x] != 0) page_test=1; // is occupied
   }
  if (page_test==1) page++;       // check the next page
  }

 z10=temp_int/10;         
 z1=temp_int-10*z10;

 PORTD=0xB8 | page;               // set page 
 Write_CData(2); 

 PORTD  =  0xC0;                  // start line=0 
 Write_CData(2);    

 for (x=111; x<128; x++)          // Clear the entire number range in the display
 {
  ym = 0x40 | (x-64);             // x=64..127, ym=0x40..0x7F            
  PORTD=ym;                       // Load column address
  Write_CData(2);       
  PORTD=0;                        // blank
  Write_DData(2);                 // write display RAM data 
  }  
 // ---------------------------------------------------------------------------
 for (x=123; x<128; x++)          // die ones digit 
 {    
  ym = 0x40 | (x-64);                      
  PORTD=ym;                        
  Write_CData(2);      
  PORTD=num[z1][x-123];            
  Write_DData(2);             
  }    
 // ---------------------------------------------------------------------------
 if (z10>0)                       // die tens digit (kill leading zero)
 {
  for (x=117; x<122; x++) 
  {   
   ym = 0x40 | (x-64);            
   PORTD=ym;                         
   Write_CData(2);      
   PORTD=num[z10][x-117];  
   Write_DData(2);              
   } 
  }
 // ---------------------------------------------------------------------------
 if (temp<=-0.5)                      // at minus temperature
 //if (temp<=-1)                      // at minus temperature
 {
  if (z10>0)                      // Temperature value two digits
  {
   for (x=111; x<116; x++) 
   {    
    ym = 0x40 | (x-64);           
    PORTD=ym;                         
    Write_CData(2);      
    PORTD=num[10][x-111];         // minus sign left of tens digit
    Write_DData(2);               
    } 
   }

  else                            // temperature value one digit
  {
   for (x=117; x<122; x++) 
   {    
    ym = 0x40 | (x-64);           
    PORTD=ym;                         
    Write_CData(2);      
    PORTD=num[10][x-117];         // minus sign instead of tens digit
    Write_DData(2);               
    } 
   }
  }
 // ---------------------------------------------------------------------------           
  PORTD  =  0x3F;                 // Display on (repetition is necessary !)
  Write_CData(2);                 // write control data  
 }
 
// ****************************************************************************

void check_TWINT (void) // wait until TWINT is set
{
 while ((TWCR&0x80)==0);
 //while (!(TWCR & (1<<TWINT))); // alternative notation
 }


void write_to_config(void) // load configuration register
{
 TWCR=0xA4; // start condition
 check_TWINT();
 
 TWDR=0x90; // slave address and write
 TWCR=0x84; // start transmission
 check_TWINT();
  
 TWDR=0xAC; // master sends access config command
 TWCR=0x84; 
 check_TWINT();

 TWDR=0x0C; // 12-bit, continuous mode
 TWCR=0x84; 
 check_TWINT();
 
 TWCR=0x94; // stop condition
 }

void start_convert (void) //  start temperature conversion
{
 TWCR=0xA4; // start condition
 check_TWINT();

 TWDR=0x90; // slave address and write
 TWCR=0x84; // start transmission
 check_TWINT();
 
 TWDR=0x51; // start convert command
 TWCR=0x84; 
 check_TWINT();

 TWCR=0x94; // stop condition
 }


void read_temp (void) // Read temperature from DS1631
{  
 TWCR=0xA4; // start condition
 check_TWINT();

 TWDR=0x90; // slave address and write
 TWCR=0x84; // start transmission
 check_TWINT();
 
 TWDR=0xAA; // read temperature 
 TWCR=0x84; 
 check_TWINT();

 TWCR=0xA4; // repeat start condition
 check_TWINT();

 TWDR=0x91; // slave address and read
 TWCR=0x84; // start transmission
 check_TWINT();

 TWCR=0x84; // read Highbyte
 check_TWINT();
 temph=TWDR;

 TWCR=0x84; // read Lowbyte
 check_TWINT();
 templ=TWDR;

 TWCR=0x94; // stop condition
 }
 

void cal_temp (void) // calculate the temperature value as float
{   
 if ((temph&0x80)==0) temp=0; else temp=-128;
 if ((temph&0x40)!=0) temp=temp+64;
 if ((temph&0x20)!=0) temp=temp+32;
 if ((temph&0x10)!=0) temp=temp+16;
 if ((temph&0x08)!=0) temp=temp+8;
 if ((temph&0x04)!=0) temp=temp+4;
 if ((temph&0x02)!=0) temp=temp+2;
 if ((temph&0x01)!=0) temp=temp+1;
 
 if ((templ&0x80)!=0) temp=temp+0.5;
 if ((templ&0x40)!=0) temp=temp+0.25;
 if ((templ&0x20)!=0) temp=temp+0.125;
 if ((templ&0x10)!=0) temp=temp+0.0625;
 }


void cal_measval (void) // Determine measured value (-23°C => 63, +40°C => 0)
{
 if (temp<0)
 {
  temp_int=(0.5-temp);  // Round temperature value
  measval=40+temp_int;
  }

 else // temperature >=0
 {
  temp_int=temp+0.5;    // Round temperature value
  measval=40-temp_int;
  }
 }

// *********************************************************************************************************

void setup()
{
 DDRB=0xFF;         // all pins of PORTB push-pull  
 DDRD=0xFF;         // all pins of PORTD push-pull 

 TWSR=0xFA;         // two wire prescaler :16
 TWBR=0x05;         // divider :5 => SCL-clock 91kHz
 
 write_to_config(); // load configuration register of thermometer DS1631
 start_convert();   // start temperature conversion
 Ini_Gdisp();       // initilize the grafic display 128x64
 clear_dispreg();   // clear display register
 set_ytics();       // set y tics
 set_ylabels();     // set y labels
 
 load_gdisp();      // load the complete graphic display with the frame
 delay(1000);       // wait until DS1631 is ready  

 TCCR1A=0x00;       // normal mode, ATTENTION: Initial Value was overwritten !
 TCCR1B=0x04;       // Timer 1 clock prescaler /256
 TIMSK1=0x01;       // Timer 1 overflow interrupt enabled
 SREG=0x80;         // global interrupt enabled
}
 
void loop()
{      
 if (meas_flag==1)      // measure and display a temperature value  
 {
  meas_flag=0;          // Start waiting time until next measurement               
 
  read_temp();          // read the temperature from DS1631
  cal_temp();           // alculate the temperature
  temp2=temp;           // save temp
 
  if (temp>40)  temp=0; // Overrange => Redirect to zero line
  if (temp<-23) temp=0; //                  "
  cal_measval();        // calculate the y-value of the display
  set_meas(measval);    // set a new measured value  
  load_gdisp();         // load the complete graphic display with the display register  

  temp=temp2;           // restore temp    
  cal_measval();           // round up unlimited temperature again 
  load_num();           // show the current temperature as a number in the display
  //delay(500);           // 60s/cycle         
  //delay(5000);          // 9min/cycle 
  //delay(32860);         // 1h/cycle
  //delay(789000);        // 24h/cycle
  shift_left();         // move all old measured values to the left      
  } // End of current measurement 
 }  


ISR(TIMER1_OVF_vect) 
{  
 TCNT1=0x0BDC;       // 1s*16MHz/256=62500counts up to 65536 => start value 3036
 count_sec++;        // Seconds counter
 //if (count_sec==3) // start the next measurement (test only)
 if (count_sec==800) // Measuring interval 24*3600s/108=800s is expired => start next measurement
 {
  count_sec=0;
  meas_flag=1;       // trigger the measurement      
  }
 }
